
/* TODO:
 * Page Boundaries Disassembly
 * implement Transfer Memory Command
 * TRACE ?!
 */

/* changelog:
 * 03.03.2007: fixed dir output from command line
 * 02.03.2007: added Makefile
 * 02.03.2007: added support for filtered directory
 * 02.03.2007: changd tokenizer, changes in string treatment, strings in quotation marks etc...
 * 01.03.2007: fixed wrong memory end address display at PUT
 * 01.03.2007: added libreadline functions for Linux
 * 01.03.2007: added feature: give cmomands at command-line
 * 28.02./ 01.03.2007: rewrote network stuff for also compiling for WIN32
 * some minor changes...
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include </home/hannenz/c64/netmon/netmon.h>

#ifdef _WIN32
#include <winsock.h>
#include <windows.h>
#include <io.h>
#else
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <readline/readline.h>
#include <readline/history.h>
extern int h_errno;

#endif

/* global vars */

#ifdef _WIN32
SOCKET fd;
#else
int fd;
#endif

int cmd,old_cmd,tokens;
char string[100];
char *token[256];
unsigned char sendbuffer[512],filebuffer[512];
int next[4096];

/* kmp search algorithm */
void init_next(char *pattern,int m){
	int i,j;
	i = 0;
	j = next[0] = -1;
	while (i < m){
		while( j > -1 && pattern[i] != pattern[j])
			j = next[j];
		++i;++j;
		(pattern[i] == pattern[j]) ? (next[i] = next[j]) : (next[i] = j);
	}
}

int kmp_search(char *pattern, char *string, unsigned size, unsigned base){
	int i = 0, j = 0;
	int m = strlen(pattern);
	int n = size;
	
	init_next (pattern,m);
	
	while(i < n){
		while (j>=0 && string[i] != pattern[j])
			j = next[j];
		++i;
		++j;
		if (j == m){
			printf("%04x\n",base + i-j);
			j = next[j];
		}
	}
	return -1;
}

/* PetASCII <--> ASCII Converting */
char pet2asc(char c){
	c &= 0x7f;
	if (c == 13 || c == 10 || c == 9) return c;
	if (c < 32) return '.';
	if (c < 65) return c;
	if (c < 91) return c + 32;
	return c - 32;
}

int string2pet(char *s){
	int i,len = strlen(s);
	for (i=0;i<len;++i)
		s[i] = pet2asc(s[i]);
	return len;
}

/* parse input string into tokens */
int tokenize(char *s){
	int i = 0;
	
	while (1){
		while (*s == ' ') ++s;		//skip leading spaces

		if (*s == '\0') return i;	//end of input string? then return number of tokens read

		if (*s == '"'){
			token[i++] = s++;	//set the token to the first char after the quot.mark
			while (*s != '"')	//search the next quot.mark or end of input string
				if (*s++ == '\0') return i;
			*s++ = '\0';		//replace the quot.mark by delimiter, advance in input string and loop
			continue;
		}

		token[i++] = s;			//else we have a token
		while (*s != ' ')		//read the token until next space or end of input string
			if (*s++ == '\0') return i;
		*s++ = '\0';
	}
}

/* send a request for memory transfer of size bytes from address base to c64 */
void send_memory_request(unsigned base,unsigned size){
	sendbuffer[0] = GET;
	sendbuffer[1] = size;
	sendbuffer[3] = (base & 0x0000ff);
	sendbuffer[4] = (base & 0x00ff00) >> 8;
	sendbuffer[5] = (base & 0xff0000) >> 16;
	send (fd,sendbuffer,6,0);
}

/* send request for directory dislplay of current device and print to stdout*/
void directory(char *s){
	int i,len = string2pet(s);

	sendbuffer[0] = DIR;
	sendbuffer[1] = len;
	memcpy(&sendbuffer[2],s,len);
	send(fd,sendbuffer,len+2,0);

	do{
		len = recv(fd,filebuffer,sizeof(filebuffer),0);
		if (len){
			for(i=0;i<len;++i)
				putchar(pet2asc(filebuffer[i]));
		}
	}while(len == 128);
}

void read_file(void){
	int fnlen,len,i;
	char *localfile,*remotefile;
	FILE *f;
	
	if (tokens < 2 || tokens > 3) puts(helptext[cmd]);
	else {
		if (token[1][0] == '"')
			remotefile = &token[1][1];
		else
			remotefile = &token[1][0];
		if (tokens == 2)
			localfile = remotefile;
		else
			localfile = token[2];
		if ((f = fopen(localfile,"wb"))==NULL){
			perror(localfile);
			return;
		}
		string2pet(remotefile);
		fnlen = strlen(remotefile);
		
		sendbuffer[0] = READ;
		sendbuffer[1] = fnlen;
		memcpy(&sendbuffer[2],remotefile,fnlen);
		send(fd,sendbuffer,fnlen+2,0);
		do{
			len = recv(fd,filebuffer,sizeof(filebuffer),0);
			if (len)
				fwrite (filebuffer,1,len,f);
		}while(len == 128);
		fclose(f);
	}
}

void write_file(void){
	int fnlen,len,i;
	char *localfile,*remotefile;
	FILE *f;

	if (tokens < 2 || tokens > 3) puts(helptext[cmd]);
	else {
		localfile = token[1];
		if (tokens == 2)
			remotefile = localfile;
		else{
			if (token[2][0] == '"')
				remotefile = &token[2][1];
			else
				remotefile = &token[2][0];
		}
		if ((f = fopen(localfile,"rb")) == NULL){
			perror(localfile);
			return;
		}
		string2pet(remotefile);
		fnlen = strlen(remotefile);

		/* first we send request to open the write file */
		sendbuffer[0] = WRITE;
		sendbuffer[1] = fnlen;
		memcpy(&sendbuffer[2],remotefile,fnlen);
		send (fd,sendbuffer,fnlen+2,0);
	
		/* and wait for OK from c64 */
		len = recv(fd,sendbuffer,sizeof(sendbuffer),0);
		if (sendbuffer[0]){
			puts("an error occured...");
			fclose(f);
			return;
		}
		
		/* now we transfer the file */
		while (!feof(f)){
			len = fread(filebuffer,1,128,f);
			sendbuffer[0] = XWRITE;
			sendbuffer[1] = len;
			memcpy(&sendbuffer[2],filebuffer,len);
			send(fd,sendbuffer,len+2,0);
			/* and wait after each packet for a (any) response from c64 */
			len = recv(fd,sendbuffer,sizeof(sendbuffer),0);
		}
		/* at last we send a OK, so the c64 can close the file */
		sendbuffer[0] = XDONE;
		send(fd,sendbuffer,1,0);
		fclose(f);
	}
}

/* transfer a file from pc to c64 memory */
void put(void){
	FILE *f;
	char *filename;
	unsigned load_addr,len,bytes = 0;
	char hi,lo;

	if (tokens < 2 || tokens > 3) puts(helptext[cmd]);
	else{
		filename = token[1];
		
		if ((f = fopen(filename,"rb")) == NULL){
			perror(filename);
			return;
		}
		/* if no load addres is given read it from file, else use the given one */
		if (tokens == 3) sscanf(token[2],"%x",&load_addr);
		else load_addr = fgetc(f) + fgetc(f) * 256;
		
		printf("sending %s to $%04x ",filename,load_addr);


		/* now send the file's data packet-wise */
		while(!feof(f)){
			len = fread(filebuffer,1,128,f);	//read len bytes data from file
			sendbuffer[0] = PUT;			// setup "put" request...
			sendbuffer[1] = len;
			sendbuffer[2] = 0;			//unused...
			sendbuffer[3] = (load_addr & 0x0000ff);
			sendbuffer[4] = (load_addr & 0x00ff00) >> 8;
			sendbuffer[5] = (load_addr & 0xff0000) >> 16;
			memcpy(&sendbuffer[6],filebuffer,len);	//copy the actual data into the packet
			send(fd,sendbuffer,len+6,0);		//and send it...
			load_addr += len;
			bytes += len;
			len = recv(fd,filebuffer,sizeof(filebuffer),0);	// wait for response and loop
		}
		
		printf("- $%04x\n",load_addr);
		fclose(f);
		printf("%d bytes transferred\n",bytes);
	}
}

/* get memory from c64 and save to a file */
void get(void){
	unsigned from, to;
	char *filename;
	FILE *f;
	int len;
	
	if (tokens != 4) puts(helptext[cmd]);
	else{
		sscanf(token[1],"%x",&from);
		sscanf(token[2],"%x",&to);
		filename = token[3];
		if ((f = fopen(filename,"wb")) == NULL){
			perror(filename);
			return;
		}
		fputc(from & 0xff,f);
		fputc(from >> 8  ,f);
		while (from < to){
			if ((len = to - from) >= 128)
				len = 128;
			send_memory_request(from,len);
			len = recv(fd,filebuffer,sizeof(filebuffer),0);
			fwrite(filebuffer,len,1,f);
			from += len;
		}
		fclose(f);
	}
}	


/* make c64 jump to the given address */
void jump(void){
	unsigned jump_addr;
	
	if (tokens != 2) puts(helptext[cmd]);
	else{
		sscanf(token[1],"%x",&jump_addr);
		sendbuffer[0] = JUMP;
		sendbuffer[1] = (jump_addr & 0xff);
		sendbuffer[2] = (jump_addr & 0xff00) >> 8;
		send(fd,sendbuffer,3,0);
	}
}

void call(void){
	unsigned sys_addr;
	unsigned a,x,y,s,i;
	a = x = y = s = 0;
		
	if (tokens <2 || tokens > 6) puts(helptext[cmd]);
	else{
		sscanf(token[1],"%x",&sys_addr);
		
		if (tokens > 2)
			sscanf(token[2],"%x",&a);
		if (tokens > 3)
			sscanf(token[3],"%x",&x);
		if (tokens > 4)
			sscanf(token[4],"%x",&y);
		if (tokens > 5)
			sscanf(token[5],"%x",&s);
		
		sendbuffer[0] = SYS;
		sendbuffer[1] = sys_addr & 0xff;
		sendbuffer[2] = sys_addr >> 8;
		sendbuffer[3] = a;
		sendbuffer[4] = x;
		sendbuffer[5] = y;
		sendbuffer[6] = s;
		send(fd,sendbuffer,7,0);

		recv(fd,sendbuffer,sizeof(sendbuffer),0);
		a = sendbuffer[0];
		x = sendbuffer[1];
		y = sendbuffer[2];
		s = sendbuffer[3];
		
		printf("returned from $%04x with\n.A .X .Y SR NV-BDIZC\n%02x %02x %02x %02x ",sys_addr,a,x,y,s);
		for (i=0;i<8;++i){
			s & 0x80 ? putchar('1') : putchar('0');
			s <<= 1;
			}
		putchar('\n');
	}
}


/* detect address mode for a given opcode; used for disassembly */
int mode (unsigned char opcode){
		
	if (opcode == 0x20) return ABS;
	if (opcode < 0x80){
		if (opcode == 0x24) return ZP;
		if (opcode == 0x6c) return IND;
		opcode &= 0x1f;
			if (opcode == 0) return IMPL;
			if (opcode == 1) return INDX;
			return table1[opcode];
	}
	else{
		if (opcode == 0x80) return IMPL;
		if (opcode == 0xa2) return IMM;
		if (opcode == 0x89) return IMPL;
		if (opcode == 0xd4) return IMPL;
		if (opcode == 0xf4) return IMPL;
		if (opcode == 0xd6) return ZPX;
		if (opcode == 0xf6) return ZPX;
		if (opcode == 0xbc) return ABSX;
		if (opcode == 0x9e) return IMPL;
		if (opcode == 0xbe) return ABSY;
		opcode &= 0x1f;
		return table2[opcode];
	}
}

/* get data from c64, disassemble and print to stdout */
/* this one is a late night hack, could suerly be optimized...! but it works.*/
void disassembly (void) {
	static unsigned from = 0;
	unsigned to;
	int len,i,j,length;
	unsigned char *p;
	unsigned char opcode;
	
	if (tokens > 3) puts (helptext[cmd]);
	else {
		if (tokens > 1)
			sscanf(token[1],"%x",&from);
		to = from + 0x30;
		if (tokens > 2)
			sscanf(token[2],"%x",&to);

		while (from < to){
			if ((len = to - from) >= 128)
				len = 128;
			send_memory_request(from,len);
			len = recv(fd, filebuffer, sizeof(filebuffer),0);

			p = &filebuffer[0];
			for (i = 0; i < len; i += j){
				printf("%04x   ",from + i);

				switch (mode(*p)){
					case IMM:
						printf("%02x %02x    ",*p,*(p+1));
						printf("  %s #$%02x",mnemo[*p],*(p+1));
						j = 2;
						break;
					case ABS:
						printf("%02x %02x %02x ",*p,*(p+1),*(p+2));
						printf("  %s $%04x",mnemo[*p],*(p+1) + *(p+2) * 256);
						j = 3;
						break;
					case ABSX:
						printf("%02x %02x %02x ",*p,*(p+1),*(p+2));				
						printf("  %s $%04x,x",mnemo[*p],*(p+1) + *(p+2) * 256);
						j = 3;
						break;
					case ABSY:
						printf("%02x %02x %02x ",*p,*(p+1),*(p+2));				
						printf("  %s $%04x,y",mnemo[*p],*(p+1) + *(p+2) * 256);
						j = 3;
						break;
					case ZP:						
						printf("%02x %02x    ",*p,*(p+1));
						printf("  %s $%02x",mnemo[*p],*(p+1));
						j = 2;
						break;
					case ZPX:
						printf("%02x %02x    ",*p,*(p+1));
						printf("  %s $%02x,x",mnemo[*p],*(p+2));
						j = 2;
						break;
					case ZPY:
						printf("%02x %02x    ",*p,*(p+1));
						printf("  %s $%02x,y",mnemo[*p],*(p+1));
						j = 2;
						break;
					case AKKU:
						printf("%02x       ",*p);
						printf("  %s a",mnemo[*p]);
						j = 1;
						break;
					case INDX:
						printf("%02x %02x    ",*p,*(p+1));
						printf("  %s ($%02x,x)",mnemo[*p],*(p+1));
						j = 2;
						break;
					case INDY:
						printf("%02x %02x    ",*p,*(p+1));
						printf("  %s ($%02x),y",mnemo[*p],*(p+1));
						j = 2;
						break;
					case REL:
						printf("%02x %02x    ",*p,*(p+1));
						printf("  %s $%04x",mnemo[*p],(from + i + 2) + (signed char)*(p+1));
						j = 2;
						break;
					case IND:
						printf("%02x %02x %02x ",*p,*(p+1),*(p+2));
						printf("  %s ($%04x)",mnemo[*p],*(p+1) + *(p+2) * 256);
						j = 3;
						break;
					case IMPL:
						printf("%02x       ",*p);
						printf("  %s",mnemo[*p]);
						j = 1;
				}
				putchar('\n');
				p += j;
			}
			from += len;
		}
	}
}

/* memory peek of c64 memory */
void mon(void){
	static unsigned from = 0;
	unsigned to;
	int len,i,j,col;
	char c;
	int width = 8;		//bytes per line, adjust to your preferences...!
	
	if (tokens > 3)	puts(helptext[cmd]);
	else{
		/* parse parameters: m [from [to]] */
		if (tokens > 1)
			sscanf(token[1],"%x",&from);
		to = from + 0x100;
		if (tokens > 2)
			sscanf(token[2],"%x",&to);
		
		/* request memory from c64 */	
		while (from < to){
			if ((len = to - from) >= 128)
				len = 128;

			send_memory_request(from,len);
			len = recv(fd, filebuffer, sizeof(filebuffer),0);
		
			/* display it to stdout */
			for(i=0;i<len;i+=j){
				printf("\n%04x : ",from + i);
				col = 7;
				for(j=0;j<width;++j){
					if (i+j >= len) break;
					printf("%02x ",filebuffer[i+j]);
					col += 3;
				}
				while (col++ < (7 + width*3)) putchar(' ');
				printf(" | ");
				col += 3;
				for(j=0;j<width;++j){
					if (i+j >= len) break;				
					putchar(isprint(filebuffer[i+j])?filebuffer[i+j]:'.');
					++col;
				}
				while(col++ < (10+width*3+width)) putchar(' ');
				printf(" | ");
				for(j=0;j<width;++j){
					if (i+j >= len) break;
					c = pet2asc(filebuffer[i+j]);
					putchar(isprint(c)?c:'.');
				}
			}
			from += len;
		}
		puts("\n");
	}
}

/* petascii dump */
void interrogate(void){
	static unsigned from;
	unsigned to;
	int len,i,j;
	char c;
	
	if (tokens > 3)	puts(helptext[cmd]);
	else{
		if (tokens > 1)
			sscanf(token[1],"%x",&from);
		to = from + 0x400;
		if (tokens > 2)
			sscanf(token[2],"%x",&to);
			
		while (from < to){
			if ((len = to - from) >= 128)
				len = 128;

			send_memory_request(from,len);
			len = recv(fd, filebuffer, sizeof(filebuffer),0);

			for (i=0;i<len;i++){
				if ((i & 63)==0)
					printf("\n%04x : ",from + i);
				c = pet2asc(filebuffer[i]);
				putchar(isprint(c)?c:'.');
			}
			from += len;
		}
		printf("\n");
	}
}

/* set/ tell device number of current device */
void device(char *s){
	unsigned device;
	
	device = atoi(s);

	if (device == 0){
		sendbuffer[0] = TELLDEV;
		send(fd,sendbuffer,1,0);
		recv(fd,sendbuffer,sizeof(sendbuffer),0);
		printf("current device is #%d\n",sendbuffer[0]);
	}
	else{
		sendbuffer[0] = SETDEV;
		sendbuffer[1] = device;
		send(fd,sendbuffer,2,0);
	}
}

/* send disk command/ read command channel */
void disk_command(char *s){
	unsigned len;

	string2pet(s);
	len = strlen(s);
	sendbuffer[0] = DISKCOMMAND;
	sendbuffer[1] = len;
	memcpy(&sendbuffer[2],s,len);
	send(fd,sendbuffer,len+2,0);
	len = recv(fd,sendbuffer,sizeof(sendbuffer),0);
	puts(sendbuffer);
}

//send request to fill memory
void fill(void){
	unsigned from,to,value;

	if (tokens != 4) puts(helptext[cmd]);
	else{
		sscanf(token[1],"%x",&from);
		sscanf(token[2],"%x",&to);
		sscanf(token[3],"%x",&value);
		sendbuffer[0] = FILL;
		sendbuffer[1] = value;
		sendbuffer[2] = (from & 0xff);
		sendbuffer[3] = (from >> 8);
		sendbuffer[4] = (to & 0xff);
		sendbuffer[5] = (to >> 8);
		send (fd,sendbuffer,6,0);
	}
}

//get memory and hunt for occurances of search pattern
void hunt(void){
	int i,len,x;
	char pattern[128],*p,*mem;
	unsigned from,to,size;
		
	if (tokens < 4) puts(helptext[cmd]);
	else {
		sscanf(token[1],"%x",&from);
		sscanf(token[2],"%x",&to);

		if (from >= to) return;
		size = to - from;
		if ((mem = (char*)malloc(size))==NULL){
			perror(NULL);
			return;
		}
		
		if (token[3][0] == '"'){
			strncpy(pattern,&token[3][1],sizeof(pattern));
			string2pet(pattern);
		}
		else{
			len = 0;
			for (i=3;i<tokens;++i)
				sscanf(token[i],"%x",&pattern[len++]);
			pattern[len++] = 0;
		}

		for (i = 0; i < size ;i += len){
			if ((len = size - i) >= 128)
				len = 128;
			send_memory_request(from + i,len);
			len = recv(fd,filebuffer,sizeof(filebuffer),0);
			memcpy(mem + i,filebuffer,len);
		}
		kmp_search(pattern,mem,size,from);
		free(mem);		
	}
}

//get two memory areas and compare them, print differences to stdout
void compare(void){
	unsigned from,to,with,size,i,len,rlen,missmatch;
	unsigned char *buf1,*buf2,*b1,*b2;

	if (tokens != 4) puts(helptext[cmd]);
	else{
		sscanf(token[1],"%x",&from);
		sscanf(token[2],"%x",&to);
		sscanf(token[3],"%x",&with);
		if (from >= to) return;
		size = to - from;
		if ((buf1 = (char*)malloc(size)) == NULL){
			perror(NULL);
			return;
		}
		if ((buf2 = (char*)malloc(size)) == NULL){
			perror(NULL);
			return;
		}
		for (i = 0; i < size; i += len){
			if ((len = size - i) >= 128)
				len = 128;
			send_memory_request(from + i,len);
			rlen = recv(fd,filebuffer,sizeof(filebuffer),0);
			memcpy(buf1 + i,filebuffer,rlen);
			
			send_memory_request(with + i,len);
			rlen = recv(fd,filebuffer,sizeof(filebuffer),0);
			memcpy(buf2 + i,filebuffer,rlen);
		}
		missmatch = 0;
		b1 = buf1; b2 = buf2;
		for (i=0;i<size;++i){
			if (*b1 != *b2){
				printf("%04x %02x - %04x %02x\n",from+i,*b1,with+i,*b2);
				++missmatch;
			}
			++b1; ++b2;
		}
		if (missmatch) printf("%d ",missmatch);
		else printf("no ");
		puts("mismatches found.");
		free(buf1); free(buf2);
	}
}

/* send memory write request */
void poke(void){
	unsigned address,len,i;
	unsigned char values[128],*p;
	
	if (tokens < 3)	puts(helptext[cmd]);
	else{
		//string...?!
		if (token[2][0] == '"'){
			strncpy(values,&token[2][1],sizeof(values));
			string2pet(values);
			len = strlen(values);
		}
		//values
		else{
			len = 0;
			for (i=2;i<tokens;++i)
				sscanf(token[i],"%x",&values[len++]);
		}

		sscanf(token[1],"%x",&address);
		
		sendbuffer[0] = POKE;
		sendbuffer[1] = len;
		sendbuffer[2] = address & 0xff;
		sendbuffer[3] = address >> 8;
		memcpy(&sendbuffer[4],values,len);
		send(fd,sendbuffer,len+4,0);
	}
}

/* help that poor user...! */
void help(void){
	int i;
	
	if (tokens == 1){
		for (i=0;i < COMMANDS;++i)
			puts(helptext[i]);
	}
	else{
		for (i = 0;i < COMMANDS; ++i)
			if (strcmp(token[1],command_table[i])==0) break;
		if (i == COMMANDS) printf("no help available for %s\n",token[1]);
		else puts(helptext[i]);
	}
}

/* set ram access mode */
void setram(void){
	int access;

	if (tokens > 2) puts (helptext[cmd]);

	if (strcmp(token[0],"rom") == 0)
		access = 0x37;
	else if (strcmp(token[0],"reu") == 0)
		access = 0xff;
	else{
		if (tokens == 1)
			access = 0x34;
		else
			sscanf(token[1],"%x",&access);
	}
	sendbuffer[0] = SETRAM;
	sendbuffer[1] = access;
	send(fd,sendbuffer,2,0);
}

/* tell server status */
void status(void){
	int len;
	sendbuffer[0] = STATUS;	
	send(fd,sendbuffer,1,0);
	len = recv(fd,sendbuffer,sizeof(sendbuffer),0);
	printf("server reports:\nmemory access is $%02x",sendbuffer[0]);
	if (sendbuffer[0] & 0x80) printf("(REU)");
	printf("\ncurrent device is %d\n",sendbuffer[1]);
}


int main (int argc, char ** argv) {
	struct sockaddr_in server;
	struct hostent *host_info;
	unsigned long addr;
	char server_addr[16], server_port[6], server_ip[16];
	int i,x,len,direct_flag = 0;
	char *cline = NULL;

	/* set default values */
	old_cmd = 0;
	strcpy(server_addr, default_ip);
	strcpy(server_port,default_port);

#ifdef _WIN32
	/* Initialisiere TCP für Windows ("winsock") */
	WORD wVersionRequested;
	WSADATA wsaData;
	wVersionRequested = MAKEWORD (1, 1);
	if (WSAStartup (wVersionRequested, &wsaData) != 0){
		perror( "Fehler beim Initialisieren von Winsock");
		exit(-1);
	}
#endif

	/* parse command line for options (-ip, -p)*/
	for (i=1;i<argc;++i){
		if (strcmp (argv[i], "-ip") == 0){
			if (++i >= argc){
				puts(usage);
				exit(1);
			}
			strcpy(server_addr,argv[i]);
			continue;
		}
		else if (strcmp (argv[i] ,"--help") == 0 || strcmp (argv[i],"-h") == 0 || strcmp(argv[i],"-?") == 0){
			puts(usage);
			return 0;
		}
		else if (strcmp (argv[i],"--version") == 0 || strcmp (argv[i],"-v") == 0){
			printf("netmon version %s\n",version);
			return 0;
		}
		else if (strcmp (argv[i] ,"--command") == 0 || strcmp (argv[i],"-c") == 0){
			while (++i < argc){
				strncat(string,argv[i],sizeof(string));
				if (i != argc-1)
					strncat(string," ",sizeof(string));
			}
			direct_flag = 1;
			break;
		}
		else{
			printf("invalid option: %s\ntry %s -h\n",argv[i],argv[0]);
			exit(1);
		}
	}
	
	if ((fd = socket(AF_INET,SOCK_DGRAM,0)) < 0){
		perror("socket error!");
		exit(errno);
	}

	memset (&server,0,sizeof(server));

	if ((addr = inet_addr(server_addr)) != INADDR_NONE){
		memcpy((char*)&server.sin_addr,&addr,sizeof(addr));
	}
	else{
		host_info = gethostbyname(server_addr);
		if (NULL == host_info){
			perror("unknown server");
			exit(-1);
		}
		memcpy((char*)&server.sin_addr,host_info->h_addr,host_info->h_length);
	}
	server.sin_family=AF_INET;
	server.sin_port=htons(atoi(server_port));

	if (connect(fd, (struct sockaddr*)&server, sizeof(server)) < 0) {
		perror("connect error");
		exit(errno);
	}

	//command interpreter loop
	do{

		if (!direct_flag){
#ifdef _WIN32
			printf("\n%s",prompt);
			fgets(string,sizeof(string),stdin);
			for (i=0;i< strlen(string);++i){
				if (string[i] == '\n'){
					string[i] = '\0';
					break;
				}
			}
#else
			if (cline){
				free(cline);
				cline = NULL;
			}
			cline = readline(prompt);
			if (cline && *cline)
				add_history(cline);

			strncpy(string,cline,sizeof(string));
#endif
		}
		/* if not "m","i" or "d", ignore "empty" input */
		if (strlen(string) == 0){
			if (old_cmd > 2)
				continue;
			else cmd = old_cmd;
		}
		else {
			//treat special commands: $,#,@
			if (string[0] == '@'){
				disk_command(&string[1]);
				continue;
			}
			else if (string[0] == '$'){
				directory(string);
				continue;
			}
			else if (string[0] == '#'){
				device(&string[1]);
				continue;
			}

			// parse input into tokens
			tokens = tokenize(string);
			if (tokens == 0) continue;

			//let's look, if the first token is a known command...
			for(cmd = 0; cmd < COMMANDS; ++cmd){
				if (strcmp(token[0],command_table[cmd]) == 0) 
					break;
			}
			//not found...?! what a pity.
			if (cmd == COMMANDS){
				printf("what exactly do you mean by: '%s'\ntype 'help' to get some help\n",token[0]);
				continue;
			}
			old_cmd = cmd;
		}
		
		if (cmd == 22 || cmd == 23) break;	//"quit" or "x"
		
		// call the according function, then loop
		switch (cmd){
			case 0:	 mon(); 	break;
			case 1:  interrogate(); break;
			case 2:  disassembly(); break;
			case 3:  jump(); 	break;
			case 4:  hunt(); 	break;
			case 5:  fill(); 	break;
			case 6:  compare(); 	break;
			case 7:  put(); 	break;
			case 8:  get(); 	break;
			case 9:  poke(); 	break;
			case 13: read_file();	break;
			case 14: write_file();	break;
			case 15: sendbuffer[0] = RUN;	 send (fd,sendbuffer,1,0);	 break;
			case 16: sendbuffer[0] = BASIC;	 send (fd,sendbuffer,1,0);	 break;
			case 17: 
			case 18: 
			case 19: setram();	break;
			case 20: 
			case 21: help();	break;
			case 24: status(); 	break;
			case 25: sendbuffer[0] = INIT;	 send (fd,sendbuffer,1,0);	 break;
			case 26: call(); 	break;
			case 27: string[0] = '\0';
				 for (i=1;i<tokens;++i){
					 strcat(string,token[i]);
					 strcat(string," ");
				 }
				 system(string);
		}
	}while(direct_flag == 0);

#ifdef _WIN32
	closesocket(fd);
	WSACleanup();
#else
	close(fd);
#endif
}

